@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- encoding: utf-8 -*-
 
-# Copyright (c) 2002-2020 "Neo4j,"
+# Copyright (c) "Neo4j"
 # Neo4j Sweden AB [http://neo4j.com]
 #
 # This file is part of Neo4j.
@@ -23,7 +23,50 @@ from collections import deque
 from warnings import warn
 
 from neo4j.data import DataDehydrator
+from neo4j.exceptions import (
+    ServiceUnavailable,
+    SessionExpired,
+)
 from neo4j.work.summary import ResultSummary
+from neo4j.exceptions import ResultConsumedError
+
+
+class _ConnectionErrorHandler:
+    """
+    Wrapper class for handling connection errors.
+
+    The class will wrap each method to invoke a callback if the method raises
+    SessionExpired or ServiceUnavailable.
+    The error will be re-raised after the callback.
+    """
+
+    def __init__(self, connection, on_network_error):
+        """
+        :param connection the connection object to warp
+        :type connection Bolt
+        :param on_network_error the function to be called when a method of
+            connection raises of of the caught errors. The callback takes the
+            error as argument.
+        :type on_network_error callable
+        """
+        self._connection = connection
+        self._on_network_error = on_network_error
+
+    def __getattr__(self, item):
+        connection_attr = getattr(self._connection, item)
+        if not callable(connection_attr):
+            return connection_attr
+
+        def outer(func):
+            def inner(*args, **kwargs):
+                try:
+                    func(*args, **kwargs)
+                finally:
+                    if self._connection.defunct():
+                        self._on_network_error()
+            return inner
+
+        return outer(connection_attr)
 
 
 class Result:
@@ -32,8 +75,9 @@ class Result:
     :meth:`.Session.run` and :meth:`.Transaction.run`.
     """
 
-    def __init__(self, connection, hydrant, fetch_size, on_closed):
-        self._connection = connection
+    def __init__(self, connection, hydrant, fetch_size, on_closed,
+                 on_network_error):
+        self._connection = _ConnectionErrorHandler(connection, on_network_error)
         self._hydrant = hydrant
         self._on_closed = on_closed
         self._metadata = None
@@ -46,9 +90,12 @@ class Result:
         # states
         self._discarding = False    # discard the remainder of records
         self._attached = False      # attached to a connection
-        self._streaming = False     # there is still more records to buffer upp on the wire
-        self._has_more = False      # there is more records available to pull from the server
-        self._closed = False        # the result have been properly iterated or consumed fully
+        # there are still more response messages we wait for
+        self._streaming = False
+        # there ar more records available to pull from the server
+        self._has_more = False
+        # the result has been fully iterated or consumed
+        self._closed = False
 
     def _tx_ready_run(self, query, parameters, **kwparameters):
         # BEGIN+RUN does not carry any extra on the RUN message.
@@ -69,11 +116,6 @@ class Result:
             "server": self._connection.server_info,
         }
 
-        run_metadata = {
-            "metadata": query_metadata,
-            "timeout": query_timeout,
-        }
-
         def on_attached(metadata):
             self._metadata.update(metadata)
             self._qid = metadata.get("qid", -1)  # For auto-commit there is no qid and Bolt 3 do not support qid
@@ -101,9 +143,7 @@ class Result:
         self._attach()
 
     def _pull(self):
-
         def on_records(records):
-            self._streaming = True
             if not self._discarding:
                 self._record_buffer.extend(self._hydrant.hydrate_records(self._keys, records))
 
@@ -116,14 +156,11 @@ class Result:
             self._on_closed()
 
         def on_success(summary_metadata):
+            self._streaming = False
             has_more = summary_metadata.get("has_more")
+            self._has_more = bool(has_more)
             if has_more:
-                self._has_more = True
-                self._streaming = False
                 return
-            else:
-                self._has_more = False
-
             self._metadata.update(summary_metadata)
             self._bookmark = summary_metadata.get("bookmark")
 
@@ -135,11 +172,9 @@ class Result:
             on_failure=on_failure,
             on_summary=on_summary,
         )
+        self._streaming = True
 
     def _discard(self):
-        def on_records(records):
-            pass
-
         def on_summary():
             self._attached = False
             self._on_closed()
@@ -150,14 +185,12 @@ class Result:
             self._on_closed()
 
         def on_success(summary_metadata):
+            self._streaming = False
             has_more = summary_metadata.get("has_more")
+            self._has_more = bool(has_more)
             if has_more:
-                self._has_more = True
-                self._streaming = False
-            else:
-                self._has_more = False
-                self._discarding = False
-
+                return
+            self._discarding = False
             self._metadata.update(summary_metadata)
             self._bookmark = summary_metadata.get("bookmark")
 
@@ -165,11 +198,11 @@ class Result:
         self._connection.discard(
             n=-1,
             qid=self._qid,
-            on_records=on_records,
             on_success=on_success,
             on_failure=on_failure,
             on_summary=on_summary,
         )
+        self._streaming = True
 
     def __iter__(self):
         """Iterator returning Records.
@@ -177,38 +210,51 @@ class Result:
         :rtype: :class:`neo4j.Record`
         """
         while self._record_buffer or self._attached:
-            while self._record_buffer:
+            if self._record_buffer:
                 yield self._record_buffer.popleft()
-
-            while self._attached is True:  # _attached is set to False for _pull on_summary and _discard on_summary
-                self._connection.fetch_message()  # Receive at least one message from the server, if available.
-                if self._attached:
-                    if self._record_buffer:
-                        yield self._record_buffer.popleft()
-                    elif self._discarding and self._streaming is False:
-                        self._discard()
-                        self._connection.send_all()
-                    elif self._has_more and self._streaming is False:
-                        self._pull()
-                        self._connection.send_all()
+            elif self._streaming:
+                self._connection.fetch_message()
+            elif self._discarding:
+                self._discard()
+                self._connection.send_all()
+            elif self._has_more:
+                self._pull()
+                self._connection.send_all()
 
         self._closed = True
 
     def _attach(self):
-        """Sets the Result object in an attached state by fetching messages from the connection to the buffer.
+        """Sets the Result object in an attached state by fetching messages from
+        the connection to the buffer.
         """
         if self._closed is False:
             while self._attached is False:
                 self._connection.fetch_message()
 
-    def _buffer_all(self):
-        """Sets the Result object in an detached state by fetching all records from the connection to the buffer.
+    def _buffer(self, n=None):
+        """Try to fill `self_record_buffer` with n records.
+
+        Might end up with more records in the buffer if the fetch size makes it
+        overshoot.
+        Might ent up with fewer records in the buffer if there are not enough
+        records available.
         """
         record_buffer = deque()
         for record in self:
             record_buffer.append(record)
+            if n is not None and len(record_buffer) >= n:
+                break
         self._closed = False
-        self._record_buffer = record_buffer
+        if n is None:
+            self._record_buffer = record_buffer
+        else:
+            self._record_buffer.extend(record_buffer)
+
+    def _buffer_all(self):
+        """Sets the Result object in an detached state by fetching all records
+        from the connection to the buffer.
+        """
+        self._buffer()
 
     def _obtain_summary(self):
         """Obtain the summary of this result, buffering any remaining records.
@@ -232,7 +278,35 @@ class Result:
         return self._keys
 
     def consume(self):
-        """Consume the remainder of this result and return a ResultSummary.
+        """Consume the remainder of this result and return a :class:`neo4j.ResultSummary`.
+
+        Example::
+
+            def create_node_tx(tx, name):
+                result = tx.run("CREATE (n:ExampleNode { name: $name }) RETURN n", name=name)
+                record = result.single()
+                value = record.value()
+                info = result.consume()
+                return value, info
+
+            with driver.session() as session:
+                node_id, info = session.write_transaction(create_node_tx, "example")
+
+        Example::
+
+            def get_two_tx(tx):
+                result = tx.run("UNWIND [1,2,3,4] AS x RETURN x")
+                values = []
+                for ix, record in enumerate(result):
+                    if x > 1:
+                        break
+                    values.append(record.values())
+                info = result.consume()  # discard the remaining records if there are any
+                # use the info for logging etc.
+                return values, info
+
+            with driver.session() as session:
+                values, info = session.read_transaction(get_two_tx)
 
         :returns: The :class:`neo4j.ResultSummary` for this result
         """
@@ -253,6 +327,13 @@ class Result:
         :returns: the next :class:`neo4j.Record` or :const:`None` if none remain
         :warns: if more than one record is available
         """
+        # TODO in 5.0 replace with this code that raises an error if there's not
+        # exactly one record in the left result stream.
+        # self._buffer(2).
+        # if len(self._record_buffer) != 1:
+        #     raise SomeError("Expected exactly 1 record, found %i"
+        #                      % len(self._record_buffer))
+        # return self._record_buffer.popleft()
         records = list(self)  # TODO: exhausts the result with self.consume if there are more records.
         size = len(records)
         if size == 0:
@@ -267,16 +348,9 @@ class Result:
 
         :returns: the next :class:`.Record` or :const:`None` if none remain
         """
+        self._buffer(1)
         if self._record_buffer:
             return self._record_buffer[0]
-        if not self._attached:
-            return None
-        while self._attached:
-            self._connection.fetch_message()
-            if self._record_buffer:
-                return self._record_buffer[0]
-
-        return None
 
     def graph(self):
         """Return a :class:`neo4j.graph.Graph` instance containing all the graph objects
